@@ -22,6 +22,7 @@ import os
 import threading
 import time
 import gc
+import inspect
 import datetime
 import h5py
 import numpy as np
@@ -33,6 +34,8 @@ from . import service_driver, service_task, util
 from .. import constants
 from .. import timeout, settings as spark_settings, secret, host_hash, job_id
 from ..backend import Backend
+from ...commons.util import patch_hugginface_layer_methods
+from ...commons.constants import exit_event
 
 PETASTORM_HDFS_DRIVER = constants.PETASTORM_HDFS_DRIVER
 TOTAL_BUFFER_MEMORY_CAP_GIB = constants.TOTAL_BUFFER_MEMORY_CAP_GIB
@@ -55,7 +58,7 @@ class KerasStepCounter(tf.keras.callbacks.Callback):
 
     def on_test_batch_begin(self, batch, logs={}):
         self.counter += 1
-
+    
     def get_step_count(self):
         return self.counter
 
@@ -133,7 +136,7 @@ class SparkBackend(Backend):
 
         driver.wait_for_initial_registration(self.settings.timeout)
         if self.settings.verbose >= 2:
-            print('Initial Spark task registration is complete.')
+            print('CEREBRO => Time: {}, Initial Spark task registration is complete'.format(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")))
         task_clients = [service_task.SparkTaskClient(index,
                                                      driver.task_addresses_for_driver(index),
                                                      self.settings.key, self.settings.verbose) for index in
@@ -153,32 +156,54 @@ class SparkBackend(Backend):
         self.spark_job_group = spark_job_group
         self.workers_initialized = True
 
-    def initialize_data_loaders(self, store, dataset_idx, schema_fields):
+    def initialize_data_loaders(self, store, schema_fields):
         """
         :param store:
         :param dataset_idx:
         :param schema_fields:
         """
         if self.workers_initialized:
-            remote_store = store.to_remote(self.spark_job_group, dataset_idx)
+            remote_store = store.to_remote(self.spark_job_group, None)
             shard_count = self._num_workers()
-            _, _, _, avg_row_size = util.get_simple_meta_from_parquet(store, schema_fields, None, dataset_idx)
+            _, _, _, avg_row_size = util.get_simple_meta_from_parquet(store, schema_fields, None)
             data_readers_fn = _data_readers_fn(remote_store, shard_count, schema_fields, avg_row_size,
                                                self.settings.disk_cache_size_bytes,
                                                self.settings.data_readers_pool_type, self.settings.num_data_readers)
 
             for task_client in self.task_clients:
-                task_client.initialize_data_loaders(data_readers_fn)
+                task_client.initialize_data_loaders(store.prefix_path, data_readers_fn)
 
             self.data_loaders_initialized = False
         else:
             raise Exception('Spark tasks not initialized for Cerebro. Please run SparkBackend.initialize_workers() '
                             'first!')
 
-    def train_for_one_epoch(self, models, store, dataset_idx, feature_col, label_col, is_train=True):
-        sub_epoch_trainers = [_get_remote_trainer(model, self, store, dataset_idx, feature_col, label_col,
-                                                  self.settings.verbose) \
-                              for model in models]
+    def train_for_one_epoch(self, models, store, feature_cols, label_cols, is_train=True):
+
+        mode = "Training"
+        if not is_train:
+            mode = "Validation"
+        if self.settings.verbose >= 1:
+            print('CEREBRO => Time: {}, Starting EPOCH {}'.format(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"), mode))
+
+        sub_epoch_trainers = []
+        for model in models:
+            if type(store) == dict:
+                a_store = store[model.getRunId()]
+            else:
+                a_store = store
+            
+            if type(feature_cols) == dict:
+                a_feature_col = feature_cols[model.getRunId()]
+            else:
+                a_feature_col = feature_cols
+            
+            if type(label_cols) == dict:
+                a_label_col = label_cols[model.getRunId()]
+            else:
+                a_label_col = label_cols
+            
+            sub_epoch_trainers.append(_get_remote_trainer(model, self, a_store, None, a_feature_col, a_label_col, is_train, self.settings.verbose))
 
         model_worker_pairs = [(i, j) for i in range(len(models)) for j in range(self._num_workers())]
         # take a random ordering
@@ -191,7 +216,7 @@ class SparkBackend(Backend):
         model_results = {model.getRunId(): None for model in models}
         model_sub_epoch_steps = {model.getRunId(): None for model in models}
 
-        while len(model_worker_pairs) > 0:
+        while not exit_event.is_set() and len(model_worker_pairs) > 0:
 
             for w in range(self._num_workers()):
                 # worker idle
@@ -199,12 +224,25 @@ class SparkBackend(Backend):
                     m = _get_runnable_model(w, model_worker_pairs, model_states, is_train)
                     if m != -1:
                         # runnable model found
+                        if self.settings.verbose >= 1:
+                            print('CEREBRO => Time: {}, Scheduling Model: {}, on Worker: {}'.format(
+                                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"), models[m].getRunId(), w))
+                        
+                        if type(store) == dict:
+                            a_store = store[models[m].getRunId()]
+                        else:
+                            a_store = store
+
                         self.task_clients[w].execute_sub_epoch(
-                            fn=sub_epoch_trainers[m], train=is_train, initial_epoch=models[m].getEpochs())
+                            fn=sub_epoch_trainers[m], store_prefix_path=a_store.prefix_path, train=is_train, initial_epoch=models[m].getEpochs())
 
                         model_states[m] = True
                         worker_states[w] = True
                         model_on_worker[w] = m
+
+                        if self.settings.verbose >= 1:
+                            print('CEREBRO => Time: {}, Scheduled Model: {}, on Worker: {}'.format(
+                                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"), models[m].getRunId(), w))
                 else:
                     m = model_on_worker[w]
                     if m != -1:
@@ -231,7 +269,11 @@ class SparkBackend(Backend):
                                         model_results[run_id][k].append(res[k][0])
                                     model_sub_epoch_steps[run_id].append(steps)
 
-            time.sleep(self.settings.polling_period)
+                            if self.settings.verbose >= 1:
+                                print('CEREBRO => Time: {}, Completed Model: {}, on Worker: {}'.format(
+                                    datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"), models[m].getRunId(), w))
+
+            exit_event.wait(self.settings.polling_period)
 
         # incrementing the model epoch number
         if is_train:
@@ -245,6 +287,9 @@ class SparkBackend(Backend):
             for k in res:
                 res[k] = (np.sum([rk * steps[i] for i, rk in enumerate(res[k])]) / np.sum(steps))
 
+        if self.settings.verbose >= 2:
+            print('CEREBRO => Time: {}, Completed EPOCH {}'.format(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"), mode))        
+
         return model_results
 
     def teardown_workers(self):
@@ -265,22 +310,19 @@ class SparkBackend(Backend):
         """
         return util.get_simple_meta_from_parquet(store, label_columns + feature_columns)
 
-    def prepare_data(self, store, dataset, validation, label_columns=['label'], feature_columns=['features'],
-                     num_partitions=None, parquet_row_group_size_mb=8, dataset_idx=None):
+    def prepare_data(self, store, dataset, validation, num_partitions=None, parquet_row_group_size_mb=8, dataset_idx=None):
         """
         Prepare data by writing out into persistent storage
 
         :param store: Cerebro storage object (e.g., LocalStorage, HDFSStorage).
         :param dataset: Spark DataFrame.
         :param validation: Fraction of validation data (e.g., 0.25) or name of the DataFrame column indicating validation.
-        :param label_columns: List of label/output columns (default=['label']).
-        :param feature_columns: List of feature columns (default=['features']).
         :param num_partitions: Number of data partitions of the output. If None, will default to the current number of
          input dataset partitions.
         :param parquet_row_group_size_mb: Parquet row group size in MBs (default 8 MB) .
         :param dataset_idx: Dataset index if storing multiple datasets in the same directory.
         """
-        return util.prepare_data(self._num_workers(), store, dataset, label_columns, feature_columns, validation,
+        return util.prepare_data(self._num_workers(), store, dataset, validation,
                                  num_partitions=num_partitions, dataset_idx=dataset_idx,
                                  parquet_row_group_size_mb=parquet_row_group_size_mb, verbose=self.settings.verbose)
 
@@ -304,25 +346,33 @@ def _get_runnable_model(worker, model_worker_pairs, model_states, is_train):
     return -1
 
 
-def _get_remote_trainer(estimator, backend, store, dataset_idx, feature_columns, label_columns, verbose=0):
+def _get_remote_trainer(estimator, backend, store, dataset_idx, feature_columns, label_columns, is_train=False, verbose=0):
+    run_id = estimator.getRunId()
+    if verbose >= 2:
+        print('CEREBRO => Time: {}, Collecting data metadata for Model: {}'.format(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"), run_id))
+    
     train_rows, val_rows, metadata, avg_row_size = \
         util.get_simple_meta_from_parquet(store,
                                           schema_cols=label_columns + feature_columns,
-                                          sample_weight_col=None,
                                           dataset_idx=dataset_idx)
     estimator._check_params(metadata)
     keras_utils = estimator._get_keras_utils()
-    run_id = estimator.getRunId()
 
-    # checkpointing the model if it does not exist
+    # Checkpointing the model if it does not exist.
     if not estimator._has_checkpoint(run_id):
-        model = estimator._compile_model(keras_utils)
         remote_store = store.to_remote(run_id, dataset_idx)
+
+        if verbose >= 2:
+            print('CEREBRO => Time: {}, Checkpointing artifacts for Model: {}'.format(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"), run_id))
+
         with remote_store.get_local_output_dir() as run_output_dir:
+            model = estimator._compile_model(keras_utils)
             ckpt_file = os.path.join(run_output_dir, remote_store.checkpoint_filename)
             model.save(ckpt_file)
             remote_store.sync(run_output_dir)
 
+    if verbose >= 2:
+        print('CEREBRO => Time: {}, Initializing sub-epoch trainer for Model: {}'.format(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"), run_id))
     trainer = sub_epoch_trainer(estimator, metadata, keras_utils, run_id, dataset_idx,
                                 train_rows, val_rows, backend._num_workers())
     return trainer
@@ -334,7 +384,7 @@ def _data_readers_fn(remote_store, shard_count, schema_fields, avg_row_size, cac
 
         PETASTORM_HDFS_DRIVER = constants.PETASTORM_HDFS_DRIVER
 
-        train_reader = make_reader(remote_store.train_data_path, shuffle_row_groups=False, num_epochs=None,
+        train_reader = make_reader(remote_store.train_data_path, shuffle_row_groups=False, num_epochs=1,
                                    cur_shard=index,
                                    shard_count=shard_count,
                                    hdfs_driver=PETASTORM_HDFS_DRIVER,
@@ -346,7 +396,7 @@ def _data_readers_fn(remote_store, shard_count, schema_fields, avg_row_size, cac
                                    cache_extra_settings={'cleanup': True})
 
         if remote_store.val_data_path != '' and remote_store.val_data_path is not None:
-            val_reader = make_reader(remote_store.val_data_path, shuffle_row_groups=False, num_epochs=None,
+            val_reader = make_reader(remote_store.val_data_path, shuffle_row_groups=False, num_epochs=1,
                                      cur_shard=index,
                                      shard_count=shard_count,
                                      hdfs_driver=PETASTORM_HDFS_DRIVER,
@@ -374,11 +424,11 @@ def _make_spark_thread(spark_context, spark_job_group, driver, result_queue,
             spark_context.setJobGroup(spark_job_group,
                                       "Cerebro Spark Run",
                                       interruptOnCancel=True)
-            procs = spark_context.range(0, numSlices=settings.num_workers)
+            procs = spark_context.range(0, end=settings.num_workers, numSlices=settings.num_workers)
             # We assume that folks caring about security will enable Spark RPC
             # encryption, thus ensuring that key that is passed here remains
             # secret.
-            result = procs.mapPartitionsWithIndex(_make_mapper(driver.addresses(), settings)).collect()
+            result = procs.barrier().mapPartitions(_make_mapper(driver.addresses(), settings)).collect()
             result_queue.put(result)
         except:
             driver.notify_spark_job_failed()
@@ -390,13 +440,13 @@ def _make_spark_thread(spark_context, spark_job_group, driver, result_queue,
 
 
 def _make_mapper(driver_addresses, settings):
-    def _mapper(index, _):
+    def _mapper(p):
         try:
             # https://www.google.com/search?q=keras+model+save+resource+temporarily+unavailable&oq=keras\
             # +mode&aqs=chrome.0.69i59l2j69i57j69i59j69i60l3j69i65.3390j0j4&sourceid=chrome&ie=UTF-8
             import os
             os.environ['HDF5_USE_FILE_LOCKING'] = 'FALSE'
-
+            index = int(sum(p))
             task = service_task.SparkTaskService(index, settings.key, settings.nics)
             driver_client = service_driver.SparkDriverClient(driver_addresses, settings.key, settings.verbose)
 
@@ -418,28 +468,25 @@ def sub_epoch_trainer(estimator, metadata, keras_utils, run_id, dataset_idx, tra
     feature_columns = estimator.getFeatureCols()
     user_callbacks = estimator.getCallbacks()
     batch_size = estimator.getBatchSize()
-    sample_weight_col = estimator.getSampleWeightCol()
     custom_objects = estimator.getCustomObjects()
-    user_shuffle_buffer_size = estimator.getShufflingBufferSize()
     metrics_names = [name.__name__ if callable(name) else name for name in estimator.getMetrics()]
     user_verbose = estimator.getVerbose()
 
     # Model parameters
     input_shapes, output_shapes = estimator.get_model_shapes()
     output_names = estimator.getModel().output_names
+    input_names = estimator.getModel().input_names
 
     floatx = tf.keras.backend.floatx()
     make_dataset = keras_utils.make_dataset_fn(
-        feature_columns, label_columns, sample_weight_col, metadata,
-        input_shapes, output_shapes, output_names, batch_size)
+        feature_columns, label_columns, metadata,
+        input_shapes, output_shapes, input_names, output_names, batch_size)
     fit_sub_epoch_fn = keras_utils.fit_sub_epoch_fn()
     eval_sub_epoch_fn = keras_utils.eval_sub_epoch_fn()
     transformation_fn = estimator.getTransformationFn()
-    transformation = transformation_fn if transformation_fn else None
 
     # Utility functions
     deserialize_keras_model = _deserialize_keras_model_fn()
-    calculate_shuffle_buffer_size = _calculate_shuffle_buffer_size_fn()
     pin_gpu = _pin_gpu_fn()
 
     # Storage
@@ -450,22 +497,19 @@ def sub_epoch_trainer(estimator, metadata, keras_utils, run_id, dataset_idx, tra
 
         begin_time = time.time()
 
+        # Workaround for the issue with huggingface layers needing a python
+        # object as config (not a dict) and explicit definition of get_config method.
+        # We monkey patch the __init__ method get_config methods of such layers.
+
+        # WARNING: checked out temporarily
+        # for k in custom_objects:
+        #     if issubclass(custom_objects[k], tf.keras.layers.Layer) and inspect.getmodule(custom_objects[k]).__name__.startswith('transformers.'):
+        #         patch_hugginface_layer_methods(custom_objects[k])
+
         tf.keras.backend.set_floatx(floatx)
         pin_gpu(local_task_index)
 
-        # FIXME: Enable sub-epoch data shuffling
-        # if not user_shuffle_buffer_size:
-        #     shuffle_buffer_size = calculate_shuffle_buffer_size(
-        #         hvd, avg_row_size, train_rows / num_workers)
-        # else:
-        #     shuffle_buffer_size = user_shuffle_buffer_size
-
-        if not user_shuffle_buffer_size:
-            shuffle_buffer_size = 1024 * 3
-        else:
-            shuffle_buffer_size = user_shuffle_buffer_size
-
-        # # Verbose mode 1 will print a progress bar
+        # Verbose mode 1 will print a progress bar.
         verbose = user_verbose
 
         with remote_store.get_local_output_dir() as run_output_dir:
@@ -479,32 +523,22 @@ def sub_epoch_trainer(estimator, metadata, keras_utils, run_id, dataset_idx, tra
                 model = deserialize_keras_model(
                     remote_store.get_last_checkpoint(), lambda x: tf.keras.models.load_model(x))
 
-            steps_per_epoch = int(math.ceil(train_rows / batch_size / num_workers))
-
-            # math.ceil because if val_rows is smaller than batch_size we still get the at least
-            # one step. float(val_rows) because val_rows/batch_size evaluates to zero before
-            # math.ceil
-            validation_steps = int(math.ceil(float(val_rows) / batch_size / num_workers))
-
             schema_fields = feature_columns + label_columns
-            if sample_weight_col:
-                schema_fields.append(sample_weight_col)
 
             if is_train:
-                train_data = make_dataset(data_reader, shuffle_buffer_size, shuffle=False)
+                train_data = make_dataset(data_reader, transformation_fn)
                 initialization_time = time.time() - begin_time
                 begin_time = time.time()
-                result = fit_sub_epoch_fn(starting_epoch, model, train_data, steps_per_epoch, callbacks,
-                                          verbose).history
+                result = fit_sub_epoch_fn(starting_epoch, model, train_data, callbacks, verbose).history
                 training_time = time.time() - begin_time
                 begin_time = time.time()
                 result = {'train_' + name: result[name] for name in result}
                 model.save(ckpt_file)
             else:
-                val_data = make_dataset(data_reader, shuffle_buffer_size, shuffle=False)
+                val_data = make_dataset(data_reader, transformation_fn)
                 initialization_time = time.time() - begin_time
                 begin_time = time.time()
-                result = eval_sub_epoch_fn(starting_epoch, model, val_data, validation_steps, callbacks, verbose)
+                result = eval_sub_epoch_fn(starting_epoch, model, val_data, callbacks, verbose)
                 training_time = time.time() - begin_time
                 begin_time = time.time()
                 result = [[x] for x in result]
@@ -523,6 +557,7 @@ def sub_epoch_trainer(estimator, metadata, keras_utils, run_id, dataset_idx, tra
                       'Finalization Time: {}'.format(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                         run_id, 'TRAIN' if is_train else 'VALID', initialization_time, training_time, finalization_time))
 
+            data_reader.reset()
             return result, step_counter_callback.get_step_count()
 
     return train
@@ -531,7 +566,6 @@ def sub_epoch_trainer(estimator, metadata, keras_utils, run_id, dataset_idx, tra
 def _deserialize_keras_model_fn():
     def deserialize_keras_model(model_bytes, load_model_fn):
         """Deserialize model from byte array encoded in base 64."""
-        # model_bytes = codec.loads_base64(model_bytes)
         bio = io.BytesIO(model_bytes)
         with h5py.File(bio, 'r') as f:
             return load_model_fn(f)
@@ -539,47 +573,6 @@ def _deserialize_keras_model_fn():
     return deserialize_keras_model
 
 
-def _calculate_shuffle_buffer_size_fn():
-    def calculate_shuffle_buffer_size(hvd, avg_row_size, train_row_count_per_worker):
-        """
-        Determines the shuffling buffer size such that each worker gets at most 1GB for shuffling
-        buffer such that on a single machine, among all the workers on that machine, at most
-        memory_cap_gb GB are allocated for shuffling buffer. Also, it ensures that the buffer size
-        is identical among all the workers.
-
-        example 1:
-        memory_cap_gb = 4
-        machine1: 8 workers
-        machine2: 3 workers
-        shuffle_buffer_size = 0.5 GB
-
-        example 2:
-        memory_cap_gb = 4
-            machine1: 2 workers
-            machine2: 3 workers
-        shuffle_buffer_size = 1 GB
-
-        example 3:
-        memory_cap_gb = 4
-            machine1: 2 workers
-            machine2: 8 workers
-            machine3: 5 workers
-        shuffle_buffer_size = 0.5 GB
-        """
-        local_size = hvd.local_size()
-        local_sizes = hvd.allgather([local_size])
-        max_local_size = max(local_sizes)
-
-        if max_local_size > TOTAL_BUFFER_MEMORY_CAP_GIB:
-            shuffle_buffer_size = TOTAL_BUFFER_MEMORY_CAP_GIB * BYTES_PER_GIB / avg_row_size / max_local_size
-        else:
-            shuffle_buffer_size = BYTES_PER_GIB / avg_row_size
-
-        return int(min(shuffle_buffer_size, train_row_count_per_worker))
-
-    return calculate_shuffle_buffer_size
-
-
 def _pin_gpu_fn():
     def fn(local_task_index):
         gpus = tf.config.experimental.list_physical_devices('GPU')
@@ -592,8 +585,15 @@ def _pin_gpu_fn():
 
 
 def _pin_cpu_fn():
-    def fn():
+    # def fn():
+    #     os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
+    #     tf.config.threading.set_inter_op_parallelism_threads(1)
+    #     tf.config.threading.set_intra_op_parallelism_threads(1)
+    def fn(tf, keras):
         os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
+        from tensorflow.python.eager import context
+        context._context = None
+        context._create_context()
         tf.config.threading.set_inter_op_parallelism_threads(1)
         tf.config.threading.set_intra_op_parallelism_threads(1)
 
