@@ -20,6 +20,7 @@ import numbers
 import time
 import os
 import json
+import inspect
 import numpy as np
 import tensorflow as tf
 
@@ -31,11 +32,9 @@ from pyspark.ml.util import DefaultParamsWriter, DefaultParamsReader
 
 from ...backend import codec
 from ...backend import spark
-
+from ...commons.util import patch_hugginface_layer_methods
 from .util import TF_KERAS, TFKerasUtil
-
 from .params import SparkEstimatorParams, SparkModelParams
-
 from ..estimator import CerebroEstimator, CerebroModel
 
 import threading
@@ -169,7 +168,7 @@ class SparkEstimator(PySparkEstimator, SparkEstimatorParams, SparkEstimatorParam
                      CerebroEstimator):
     """Cerebro Spark Estimator for fitting Keras model to a DataFrame.
 
-    Supports ``tf.keras >= 2.1``.
+    Supports ``tf.keras >= 2.2``.
 
     Args:
         model: Keras model to train.
@@ -179,23 +178,13 @@ class SparkEstimator(PySparkEstimator, SparkEstimatorParams, SparkEstimatorParam
         loss: Keras loss or list of losses.
         batch_size: Number of rows from the DataFrame per batch.
         loss_weights: (Optional) List of float weight values to assign each loss.
-        sample_weight_col:(Optional) Column indicating the weight of each sample.
         metrics: (Optional) List of Keras metrics to record.
         callbacks: (Optional) List of Keras callbacks.
+        transformation_fn: (Optional) Function that takes a TensorFlow Dataset as its parameter
+                       and returns a modified Dataset that is then fed into the
+                       train or validation step. This transformation is applied before batching.
     """
-
-    # TODO
-    # shuffle_buffer_size: (Optional) Size of in-memory shuffle buffer in rows. Allocating a larger buffer size
-    #                      increases randomness of shuffling at the cost of more host memory. Defaults to estimating
-    #                      with an assumption of 4GB of memory per host.
-    # transformation_fn: (Optional) Function that takes a row as its parameter
-    #                    and returns a modified row that is then fed into the
-    #                    train or validation step. This transformation is
-    #                    applied after batching. See Petastorm TransformSpec
-    #                    for more details. Note that this fucntion constructs
-    #                    another function which should perform the
-    #                    transformation.
-
+    
     custom_objects = Param(Params._dummy(), 'custom_objects', 'custom objects')
     _keras_pkg_type = Param(Params._dummy(), '_keras_pkg_type', 'keras package type')
 
@@ -207,10 +196,8 @@ class SparkEstimator(PySparkEstimator, SparkEstimatorParams, SparkEstimatorParam
                  loss=None,
                  batch_size=None,
                  loss_weights=None,
-                 sample_weight_col=None,
                  metrics=None,
                  callbacks=None,
-                 shuffle_buffer_size=None,
                  transformation_fn=None
                  ):
 
@@ -229,6 +216,10 @@ class SparkEstimator(PySparkEstimator, SparkEstimatorParams, SparkEstimatorParam
         kwargs = self._input_kwargs
         self.setParams(**kwargs)
 
+
+    def _fit(self):
+        raise NotImplementedError
+
     def _get_keras_utils(self):
         # This function determines the keras package type of the Estimator based on the passed
         # optimizer and model and updates _keras_pkg_type parameter.
@@ -432,6 +423,12 @@ class SparkModel(PySparkModel, SparkModelParams, SparkEstimatorParamsReadable, S
     def _get_floatx(self):
         return self.getOrDefault(self._floatx)
 
+
+    def _transform(self):
+        """ Only required by PySparkModel, not used at all
+        """
+        raise NotImplementedError
+
     def transform(self, df):
         """ Transforms the given DataFrame using the trained ML model
 
@@ -451,13 +448,14 @@ class SparkModel(PySparkModel, SparkModelParams, SparkEstimatorParamsReadable, S
         pin_cpu = spark.backend._pin_cpu_fn()
 
         def predict(rows):
+            import tensorflow as tf
             from pyspark import Row
             from pyspark.ml.linalg import DenseVector, SparseVector
 
             k = keras_utils.keras()
             k.backend.set_floatx(floatx)
             # Do not use GPUs for prediction, use single CPU core per task.
-            pin_cpu()
+            pin_cpu(tf, k)
 
             def load_model_fn(x):
                 with k.utils.custom_object_scope(custom_objects):
@@ -527,3 +525,15 @@ class SparkModel(PySparkModel, SparkModelParams, SparkEstimatorParamsReadable, S
             return self.getModel()
         else:
             raise Exception('Keras model is not set!')
+
+    # copied from https://github.com/apache/spark/tree/master/python/pyspark/ml/param/shared.py
+    # has been removed from pyspark.ml.param.HasOutputCol in pyspark 3.0.0
+    # added here to keep ModelParams API consistent between pyspark 2 and 3
+    # https://github.com/apache/spark/commit/b19fd487dfe307542d65391fd7b8410fa4992698#diff-3d1fb305acc7bab18e5d91f2b69018c7
+    # https://github.com/apache/spark/pull/26232
+    # https://issues.apache.org/jira/browse/SPARK-29093
+    def setOutputCols(self, value):
+        """
+        Sets the value of :py:attr:`outputCols`.
+        """
+        return self._set(outputCols=value)
